package org.andromda.samples.animalquiz;

import java.math.BigInteger;
import java.net.URI;
import java.net.URL;
import java.sql.Date;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Map;

/**
 * Stores the java mapping to schema type mappings.
 *
 * @author Chad Brandon
 */
class TypeMappings
{
    private static final Map<Class, String> mappings = new HashMap<Class, String>();
    static
    {
        mappings.put(Boolean.class, "boolean");
        mappings.put(Byte.class, "byte");
        mappings.put(byte[].class, "base64Binary");
        mappings.put(URI.class, "string");
        mappings.put(URL.class, "string");
        mappings.put(char[].class, "string");
        mappings.put(Character[].class, "string");
        mappings.put(String.class, "string");
        mappings.put(Calendar.class, "dateTime");
        mappings.put(java.util.Date.class, "dateTime");
        mappings.put(Date.class, "dateTime");
        mappings.put(Timestamp.class, "dateTime");
        mappings.put(char.class, "int");
        mappings.put(BigInteger.class, "integer");
        mappings.put(Float.class, "decimal");//TODO Is overwritten
        mappings.put(Double.class, "double");
        mappings.put(Float.class, "float");
        mappings.put(Short.class, "short");
        mappings.put(Object.class, "anyType");
    }

    /**
     * Gets the appropriate schema type from the given <code>javaType</code>.
     *
     * @param javaType the java type from which to get the schema type.
     * @return the schema type mapping.
     */
    static String getSchemaType(final Class javaType)
    {
        return mappings.get(javaType);
    }
}